// -----------------
// CubicSpline.cpp
// Jason Hughes
// Origin Systems
// Copyright 1998
// -----------------

#include "first.h"
#include "CubicSpline.h"

CubicSpline::CubicSpline(void)
{
	numSamples=0;
	T=NULL;
	Y=NULL;
	Z=NULL;
}

CubicSpline::~CubicSpline(void)
{
	delete []T;
	T=NULL;
	delete []Y;
	Y=NULL;
	delete []Z;
	Z=NULL;
}

void CubicSpline::triSolve(int samples,const real *A,real *D,const real *C,real *B) const
{
	real xmult;

	for (int i=1; i<samples; i++)
	{
		xmult=A[i-1]/D[i-1];
		D[i]=D[i]-xmult*C[i-1];
		B[i]=B[i]-xmult*B[i-1];
	}
	B[samples-1]=B[samples-1]/D[samples-1];

	for (i=1; i<samples; i++)
		B[samples-i-1]=(B[samples-i-1]-C[samples-i-1]*B[samples-i])/D[samples-i-1];
}

void CubicSpline::setSamples(int sampleCount,const Sample *sampleArray) // Numerical and Mathematical Computing, p. 155  Function ZSPL3
{
	real temp;
	real *C,*D;

	numSamples=sampleCount;
	T=NEW real[numSamples];
	Y=NEW real[numSamples];
	Z=NEW real[numSamples];
	C=NEW real[numSamples];                   // these are temporary, but I can't allocate them on the stack easily
	D=NEW real[numSamples];

	D[0]=CONST_REAL(1);
	C[0]=CONST_REAL(0);
	Z[0]=CONST_REAL(0);

	for (int i=0; i<numSamples; i++)
	{
		T[i]=sampleArray[i].x;
		Y[i]=sampleArray[i].y;
	}

	for (i=1; i<numSamples-1; i++)            // now calculate the Z array values with constants of integration C and D, then solve the tridiagonal matrix
	{
		D[i]=CONST_REAL(2)*(T[i+1]-T[i-1]);
		C[i]=T[i+1]-T[i];
		temp=(Y[i+1]-Y[i])/(T[i+1]-T[i]);
		Z[i]=CONST_REAL(6)*(temp-(Y[i]-Y[i-1])/(T[i]-T[i-1]));
	}

	D[i]=CONST_REAL(1);
	C[i-1]=CONST_REAL(0);
	Z[i]=CONST_REAL(0);

	triSolve(numSamples,C,D,C,Z);

	delete []C;                               // damn, I want this to be more efficient.
	delete []D;
}

real CubicSpline::getSample(real x) const     // Numerical and Mathematical Computing, p. 156  Function SPL3
{
	const real oo6 = RECIP(6.0f);
	real linearDelta;   // this is x - T[i], which is the delta from the start knot to the actual position on the spline
	real H;             // this is the domain between the knots where x has been found
	real ooH;
	real A,B;
	int  numSamplesMinusOne;

	numSamplesMinusOne=numSamples-1;
	DEBUG_FATAL(x<T[0] || x>T[numSamplesMinusOne],("CubicSpline::getSample() - requested %8.3f which is beyond the defined range of spline (%8.3f,%8.3f)\n",x,T[0],T[numSamplesMinusOne]));
	x=clamp(T[0],x,T[numSamplesMinusOne]);

	for (int i=1; i<numSamplesMinusOne && x>T[i]; i++)   // this finds the set of knots which enclose the passed-in x
		;
	i--;

	linearDelta=x-T[i];	

	H=T[i+1]-T[i];
	ooH=RECIP(H);                                   // this gets divided by twice, so better recip and multiply instead.

	A=linearDelta*(Z[i+1]-Z[i])*(oo6*ooH)+CONST_REAL(0.5)*Z[i];
	B=linearDelta*A+(Y[i+1]-Y[i])*ooH - H*(CONST_REAL(2)*Z[i]+Z[i+1])*oo6;
	return linearDelta*B + Y[i];
}
